use std::fmt::{Debug, Display};
use tracing_subscriber::{fmt, prelude::*, EnvFilter};

#[derive(thiserror::Error, Debug)]
#[error("No user info found")]
pub struct NoUserInfo;

#[derive(thiserror::Error, Debug)]
#[error("Handler was already used")]
pub struct HandlerUsed;

#[derive(thiserror::Error, Debug)]
pub enum InvalidCommand {
    #[error("Invalid params")]
    InvalidParams,
    #[error("Not enough bytes in the name's hash")]
    InvalidOutputLength,
    #[error("Error decoding the values: {0}")]
    NameDecodingError(#[from] base64::DecodeError),
    #[error("Unknown locale passed into callback")]
    UnknownLocale,
}

pub fn init_logger() {
    tracing_subscriber::registry()
        .with(fmt::layer().pretty())
        .with(EnvFilter::from_default_env())
        .init();
}

/// Handles an error.
/// This function must be as cheap and generic as possible and must remain sync.
/// If expensive or long action is required, a task should be spawned but not joined.
pub fn handle_error<T>(error: T)
where
    T: Debug + Display,
{
    tracing::error!(error_debug = ?error, error_message = %error);
}

pub struct ErrorHandler;

impl teloxide::error_handlers::ErrorHandler<crate::Error> for ErrorHandler {
    fn handle_error(
        self: std::sync::Arc<Self>,
        error: crate::Error,
    ) -> futures::prelude::future::BoxFuture<'static, ()> {
        handle_error(error);
        Box::pin(async {})
    }
}
